--
--  Mail_FlagsAppDelegate.applescript
--  Mail Flags
--
(*
	Mail Flags: an application to set status flags for imported e-mail
	messages in Apple Mail
	Copyright © 2011 Jakko Westerbeke <jakko@xs4all.nl>

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*)

property NSMutableArray : class "NSMutableArray"
property NSBundle : class "NSBundle" of current application
script Mail_FlagsAppDelegate
	property parent : class "NSObject"
	property mailboxes : missing value -- table view with mailboxes
	property progressBar : missing value
	property theDataSource : {}
	property buttonSetFlags : missing value
	property buttonRescan : missing value
	property menuSetFlags : missing value
	property menuRescan : missing value
	property menuLicense : missing value
	
	--Constants
	property HEADER_STATUS : "Status"
	property HEADER_X_STATUS : "X-Status"
	property HEADER_MOZILLA_STATUS : "X-Mozilla-Status"
	property SET_FLAG : true
	property UNSET_FLAG : false
	property IGNORE_FLAG : -1 -- true as integer is equal to 1 and false to 0, so using -1 as value for "ignore" works
	property MOZILLA_MESSAGE_READ : 0
	property MOZILLA_MESSAGE_REPLIED : 1
	property MOZILLA_MESSAGE_FLAGGED : 2
	property MOZILLA_MESSAGE_FORWARDED : 12
	
	
	on applicationShouldTerminate_(sender)
		return current application's NSTerminateNow
	end applicationShouldTerminate_
	
	
	on applicationWillFinishLaunching_(aNotification)
		(*
		Check whether Mail is running, and if not, warn the user that it will be started
		so he or she can decide not to proceed for whatever reason.
		*)
		set quitNow to false
		if not isMailRunning() then
			set closeApp to button returned of (display dialog (localized string "Mail Flags is about to start Apple Mail.

If you do not want Mail to be started, you should click Quit now.") buttons {localized string "Quit", localized string "Start Mail"} default button 2 with icon caution)
			if closeApp = (localized string "Quit") then
				(*
				We need to do this with a variable because giving "quit me" here will cause code
				execution to continue after that command, and so Mail will be started anyway.
				*)
				set quitNow to true
			end if
		end if
		
		if quitNow then
			quit me
		else
			set theDataSource to NSMutableArray's alloc()'s init()
			readMailboxData()
			(*
			Try to make sure Mail Flags remains at the front even after Mail starts.
			*)
			tell me
				activate
			end tell
		end if
	end applicationWillFinishLaunching_
	
	
	on awakeFromNib()
		(*
		For some reason a progress bar does not use threaded animation as default, while
		a spinning progress indicator does, so we need to set that manually else it will remain
		static while the app does its thing.
		*)
		progressBar's setUsesThreadedAnimation_(true)
	end awakeFromNib
	
	
	on enableControls(enable)
		
		buttonRescan's setEnabled_(enable)
		menuRescan's setEnabled_(enable)
		
		if mailboxes's numberOfSelectedRows() = 0 then
			buttonSetFlags's setEnabled_(false)
			menuSetFlags's setEnabled_(false)
		else
			buttonSetFlags's setEnabled_(enable)
			menuSetFlags's setEnabled_(enable)
		end if
		
		(*
		Delay necessary to make sure buttons get disabled -- without it, they look clickable
		while the app does its thing.
		*)
		delay 0.05
		
	end enableControls
	
	
	on isFlagSet(flags, flagToCheck)
		(*
		We check if a flag is set by passing its hex value to Python, letting that bit-shift it to the right,
		and then checking if it is an odd number or not (odd = bit is set, even = bit is not set).
		This is easier than writing an AppleScript routine to do the same ;)
		*)
		set flagValue to do shell script ("python -c 'print(0x" & flags & ">>" & flagToCheck & ")'")
		if flagValue mod 2 = 0 then
			return false
		else
			return true
		end if
	end isFlagSet
	
	
	on isMailRunning()
		tell application "System Events" to (name of processes) contains "Mail"
	end isMailRunning
	
	
	on numberOfRowsInTableView_(anObject)
		try
			if theDataSource's |count|() is equal to null then
				buttonSetFlags's setEnabled_(false)
				return 0
			else
				return theDataSource's |count|()
			end if
		on error
			return 0
		end try
	end numberOfRowsInTableView_
	
	
	on readMailboxData()
		
		enableControls(false)
		progressBar's startAnimation_(0)
		
		theDataSource's removeAllObjects()
		mailboxes's reloadData()
		
		set sortedMailboxes to {}
		tell application "Mail"
			set theMailboxes to mailboxes
			repeat with theMailbox in theMailboxes
				set currentMailbox to theMailbox
				set fullPath to (name of currentMailbox) as string
				repeat until (name of container of currentMailbox) as string = ""
					(*
					Make sure full pathname of each mailbox is displayed. Reason for this is
					that if we would turn theMailbox into a string, we'd only get the name of
					the actual mailbox, even though theMailbox contains the full path to it …
					*)
					set currentMailbox to container of currentMailbox
					set fullPath to ((name of currentMailbox) as string) & "/" & fullPath
				end repeat
				set end of sortedMailboxes to fullPath
			end repeat
			
			(*
			Sorting routine courtesy of http://hints.macworld.com/article.php?story=20040513173003941
			because I can't be arsed to spend several days of frustration trying to figure out
			how to do this with Objective C
			*)
			set old_delims to AppleScript's text item delimiters
			set AppleScript's text item delimiters to {ASCII character 10} -- always a linefeed
			set list_string to (sortedMailboxes as string)
			set new_string to do shell script "echo " & quoted form of list_string & " | sort -f"
			set sortedMailboxes to (paragraphs of new_string)
			set AppleScript's text item delimiters to old_delims
			
			set theData to {}
			repeat with theMailbox in sortedMailboxes
				(*
				Store name, total number of messages and number of unread messages for each mailbox
				so they can be displayed in the table.
				*)
				set messagesInTheMailbox to count messages of mailbox theMailbox
				set numberUnread to unread count of mailbox theMailbox
				set end of theData to {mailboxName:theMailbox, numberOfMessages:messagesInTheMailbox, numberUnread:numberUnread}
			end repeat
		end tell
		(*
		Stick it all into the table's data source and update the table
		*)
		theDataSource's addObjectsFromArray_(theData)
		mailboxes's reloadData()
		progressBar's stopAnimation_(0)
		
		enableControls(true)
		
	end readMailboxData
	
	
	on rescanMailboxes_(sender)
		readMailboxData()
	end rescanMailboxes_
	
	
	on setFlags()
		
		enableControls(false)
		progressBar's setIndeterminate_(false)
		progressBar's setDoubleValue_(0)
		
		(*
		Delay necessary to make sure buttons get disabled -- without it, they look clickable
		while the app does its thing.
		*)
		delay 0.05
		
		set messageCount to 0
		set selectedMailboxes to {}
		
		(*
		Work out which mailboxes are selected in the table
		*)
		repeat with aRow from 0 to count theDataSource
			if mailboxes's isRowSelected_(aRow) then
				set theRecord to theDataSource's objectAtIndex_(aRow)
				set messageCount to messageCount + (theRecord's objectForKey_("numberOfMessages") as integer)
				set end of selectedMailboxes to (theRecord's objectForKey_("mailboxName")) as string
			end if
		end repeat
		progressBar's setMaxValue_(messageCount)
		
		tell application "Mail"
			repeat with theMailbox in selectedMailboxes
				set theMessages to messages of mailbox theMailbox
				repeat with theMessage in theMessages
					progressBar's incrementBy_(1)
					set messageFrom to content of header "From" of theMessage
					set messageDate to content of header "Date" of theMessage
					--tell me to log messageFrom & " (" & messageDate & ")"
					
					(*
					Initialize/reset flags
					*)
					set messageIsFlagged to IGNORE_FLAG
					set messageIsForwarded to IGNORE_FLAG
					set messageIsRead to IGNORE_FLAG
					set messageIsRepliedTo to IGNORE_FLAG
					
					(*
					Get all headers and check them for various "Status" ones
					*)
					set allHeaders to headers of theMessage
					repeat with theHeader in allHeaders
						if name of theHeader is HEADER_STATUS or name of theHeader is HEADER_X_STATUS then
							(*
							These are common headers in mbox files and contain a string of one or more letters,
							the relevant ones to us being:
								A	Message has been replied to (answered)
								F	Message has been flagged
								R	Message has been read
							This is based on http://docs.python.org/library/mailbox.html#mailbox.mboxMessage
							*)
							set statusFlags to content of theHeader
							if statusFlags contains "A" then
								set messageIsRepliedTo to SET_FLAG
							end if
							set statusFlags to content of theHeader
							if statusFlags contains "F" then
								set messageIsFlagged to SET_FLAG
							end if
							if statusFlags contains "R" then
								set messageIsRead to SET_FLAG
							end if
						else if name of theHeader is HEADER_MOZILLA_STATUS then
							(*
							Handle flags set by Mozilla mailers (mainly, Thunderbird) by getting Mail to give us 
							the X-Mozilla-Status header, which contains Mozilla's flags for the message.
							Details of these flags can be found at
							http://mxr.mozilla.org/comm-central/source/mailnews/base/public/nsMsgMessageFlags.idl#45
					
							There are a lot more X-Mozilla-Status flags than will be converted, but most of 
							those are not relevant to Mail's message status indicators.
							*)
							set mozFlags to content of theHeader
							--tell me to log mozFlags
							if mozFlags ≠ "" then
								set messageIsFlagged to (isFlagSet(mozFlags, MOZILLA_MESSAGE_FLAGGED) of me)
								set messageIsForwarded to (isFlagSet(mozFlags, MOZILLA_MESSAGE_FORWARDED) of me)
								set messageIsRead to (isFlagSet(mozFlags, MOZILLA_MESSAGE_READ) of me)
								set messageIsRepliedTo to (isFlagSet(mozFlags, MOZILLA_MESSAGE_REPLIED) of me)
							end if
						end if
					end repeat
					
					(*
					Actually set flags.
					*)
					if messageIsFlagged ≠ IGNORE_FLAG then
						set flagged status of theMessage to (isFlagSet(mozFlags, MOZILLA_MESSAGE_FLAGGED) of me)
					end if
					if messageIsForwarded ≠ IGNORE_FLAG then
						set was forwarded of theMessage to (isFlagSet(mozFlags, MOZILLA_MESSAGE_FORWARDED) of me)
					end if
					if messageIsRead ≠ IGNORE_FLAG then
						set read status of theMessage to (isFlagSet(mozFlags, MOZILLA_MESSAGE_READ) of me)
					end if
					if messageIsRepliedTo ≠ IGNORE_FLAG then
						set was replied to of theMessage to (isFlagSet(mozFlags, MOZILLA_MESSAGE_REPLIED) of me)
					end if
					
				end repeat
			end repeat
		end tell
		
		(*
		Update table
		*)
		readMailboxData()
		
		progressBar's setIndeterminate_(true)
		enableControls(true)
		
	end setFlags
	
	
	on setMailFlags_(sender)
		if button returned of (display dialog (localized string "You are about to convert the flags on e-mail messages. This operation cannot be undone! Are you sure you want to proceed?") buttons {localized string "Cancel", localized string "Convert flags"} default button (localized string "Cancel") with icon caution) = (localized string "Convert flags") then
			setFlags()
		end if
	end setMailFlags_
	
	
	on showHelpFile_(sender)
		(*
		Shows the license for Mail Flags by telling the Finder to open the readme file from the Resources
		folder inside the app bundle.
		*)
		set helpFile to ((NSBundle's mainBundle()'s pathForResource_ofType_("Read Me", "html")) as string)
		tell application "Finder"
			open POSIX file helpFile as alias
		end tell
	end showHelpFile_
	
	
	on showLicense_(sender)
		(*
		Shows the license for Mail Flags by telling the Finder to open the license file from the Resources
		folder inside the app bundle.
		*)
		set licenseFile to ((NSBundle's mainBundle()'s resourcePath()) as string) & "/Mail Flags License.txt"
		tell application "Finder"
			open POSIX file licenseFile as alias
		end tell
	end showLicense_
	
	
	on tableView_objectValueForTableColumn_row_(anObject, aColumn, aRow)
		if theDataSource's |count|() is equal to 0 then
			buttonSetFlags's setEnabled_(false)
			return end
		end if
		set ident to aColumn's identifier
		set theRecord to theDataSource's objectAtIndex_(aRow)
		set theValue to theRecord's objectForKey_(ident)
		return theValue
	end tableView_objectValueForTableColumn_row_
	
	
	on tableView_sortDescriptorsDidChange_(anObject, oldDescriptors)
		set sortDesc to anObject's sortDescriptors()
		theDataSource's sortUsingDescriptors_(sortDesc)
		anObject's reloadData()
	end tableView_sortDescriptorsDidChange_
	
	
	on tableViewSelectionDidChange_(aNotification)
		(*
		Disable the button and menu item to convert flags if there is no selection in the table.
		*)
		if mailboxes's numberOfSelectedRows() = 0 then
			buttonSetFlags's setEnabled_(false)
			menuSetFlags's setEnabled_(false)
		else
			buttonSetFlags's setEnabled_(true)
			menuSetFlags's setEnabled_(true)
		end if
	end tableViewSelectionDidChange_
	
	
end script